
#include <windows.h>
#include <WNASPI32.h>
#include <SCSIDEFS.h>
#include <memory.h>
#include "DeCSSauth.h"

////////////////////////////////////////////////////////////////////////////////////////////////
// DeCSSauth
////////////////////////////////////////////////////////////////////////////////////////////////

typedef struct _SCSI_ADDRESS {
    unsigned long	Length;
    unsigned char	PortNumber;
    unsigned char	PathId;
    unsigned char	TargetId;
    unsigned char	Lun;
}SCSI_ADDRESS, *PSCSI_ADDRESS;

#define CTL_CODE(DeviceType,Function,Method,Access) (((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method))

#define FILE_ANY_ACCESS                 0
#define METHOD_BUFFERED                 0
#define FILE_DEVICE_CONTROLLER          0x00000004
#define IOCTL_SCSI_BASE                 FILE_DEVICE_CONTROLLER
#define IOCTL_SCSI_GET_ADDRESS          CTL_CODE(IOCTL_SCSI_BASE, 0x0406, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define DVD_SEND_KEY			0xA3
#define DVD_REPORT_KEY			0xA4
#define DVD_GET_PERFORMANCE		0xAC
#define DVD_READ_STRUCTURE		0xAD

#define copy_key(dest,src)	memcpy((dest),(src),sizeof(CSS_Key))
#define copy_chal(dest,src)	memcpy((dest),(src),sizeof(CSS_Challenge))

DeCSSauth::DeCSSauth()
{
	unsigned char	AdapterCount;
	unsigned int	ASPI32Status;

	ASPI32Status = GetASPI32SupportInfo();
	AdapterCount = (LOBYTE(LOWORD(ASPI32Status)));

	hEventEnd = CreateEvent(NULL,FALSE,FALSE,NULL);

	if ((AdapterCount!=0) && (HIBYTE(LOWORD(ASPI32Status)) == SS_COMP))
		return;
//	printf( "Error : ASPI SCSI manager not found.\n" );
//	printf( " Please add ASPI SCSI manager to your config.sys.\n" );
	return;
}

DeCSSauth::~DeCSSauth()
{
}

void DeCSSauth::setup_report_key(packet_command *pc, unsigned agid, unsigned type)
{
	memset(pc->Cmd,0,16);
	memset(pc->Buffer,0,pc->BufSize);
	pc->Cmd[0] = DVD_REPORT_KEY;
	pc->Cmd[10] = type | (agid << 6);
}

void DeCSSauth::setup_send_key(packet_command *pc, unsigned agid, unsigned type)
{
	memset(pc->Cmd,0,pc->CmdSize);
	memset(pc->Buffer,0,pc->BufSize);
	pc->Cmd[0] = DVD_SEND_KEY;
	pc->Cmd[10] = type | (agid << 6);
}

int DeCSSauth::AuthorizeDrive(char *FileName)
{
packet_command	pc;
unsigned char	buf[65536];
int				i,drive_index,index;
CSS_Key			Key1,Key2,KeyCheck;
CSS_Challenge	Key0_1 = { 9,8,7,6,5,4,3,2,1,0 };
CSS_Challenge	Key0_2 = { 0,1,2,3,4,5,6,7,8,9 };
char			FileNameFullPath[512];

	printf("Authorizing Drive... ");
	if (GetFullPathName(FileName,512,(char*)&FileNameFullPath,(char**)&i)==0) return -1;

	if (IsWindowsNTor9x()==1)
	{
		if (!GetDriveInfoNT(FileNameFullPath[0])) return -1;
	}
	else if (IsWindowsNTor9x()==2)
	{
		if (!GetDriveInfo9x(FileNameFullPath[0])) return -1;
	}
	else return -1;
	printf("HA%i,ID%i,LUN%i ... ",Host,Target,Lun);

	pc.CmdSize = 12;
	pc.BufSize = sizeof(buf);
	pc.Buffer = (unsigned char*)&buf;
	setup_report_key(&pc,0,0);
	pc.Cmd[9] = 8;
	if (!DVD_QUEUE_COMMAND(&pc)) return -1;
	i = buf[7] & 1;

	// Send Challange Key to Unit
	pc.CmdSize = 12;
	pc.Cmd[0] = DVD_SEND_KEY;
	pc.Cmd[9] = 0x10;
	pc.Cmd[10] = 0xC1;
	buf[0] = 0;
	buf[1] = 14;
	buf[2] = buf[3] = 0;
	copy_chal(&buf[4],Key0_1);
	if (!DVD_QUEUE_COMMAND(&pc)) return -1;

	// Receive Key1 from unit
	pc.CmdSize = 12;
	pc.Cmd[0] = DVD_REPORT_KEY;
	pc.Cmd[9] = 0x0C;
	pc.Cmd[10] = 0xC2;
	if (!DVD_QUEUE_COMMAND(&pc)) return -1;
	Key1[0] = buf[4+4];
	Key1[1] = buf[4+3];
	Key1[2] = buf[4+2];
	Key1[3] = buf[4+1];
	Key1[4] = buf[4+0];
	for (drive_index = -1, index=31; index >=0; index--)
	{
		CryptKey1(index,Key0_2,&KeyCheck);
		if (memcmp(&KeyCheck,&Key1, 5)==0)
		{
			drive_index = index;
			break;
		}
	}
	if (drive_index==-1) return -1;
	
	// Get Challenge Key from Unit
	pc.CmdSize = 12;
	pc.Cmd[0] = DVD_REPORT_KEY;
	pc.Cmd[9] = 0x10;
	pc.Cmd[10] = 0xC1;
	memset(&buf,0,sizeof(buf));
	buf[0] = 0;
	buf[1] = 10;
	buf[2] = buf[3] = 0;
	if (!DVD_QUEUE_COMMAND(&pc)) return -1;
	Key0_1[0] = buf[4+9];
	Key0_1[1] = buf[4+8];
	Key0_1[2] = buf[4+7];
	Key0_1[3] = buf[4+6];
	Key0_1[4] = buf[4+5];
	Key0_1[5] = buf[4+4];
	Key0_1[6] = buf[4+3];
	Key0_1[7] = buf[4+2];
	Key0_1[8] = buf[4+1];
	Key0_1[9] = buf[4+0];
	CryptKey2(drive_index,Key0_1,&Key2);

	// Send Back Key to unit
	pc.CmdSize = 12;
	pc.Cmd[0] = DVD_SEND_KEY;
	pc.Cmd[9] = 12;
	pc.Cmd[10] = 0xC3;
	buf[0] = 0;
	buf[1] = 10;
	buf[2] = buf[3] = 0;
	buf[4] = Key2[4];
	buf[5] = Key2[3];
	buf[6] = Key2[2];
	buf[7] = Key2[1];
	buf[8] = Key2[0];
	if (DVD_QUEUE_COMMAND(&pc)!=1) return -1; // wrong key or error

	CryptBusKey(drive_index,Key1,Key2,&KeyCheck);

	pc.Cmd[0] = DVD_READ_STRUCTURE;
	pc.Cmd[9] = 4;
	pc.Cmd[10] = 0xC0;
	buf[0] = 0;
	buf[1] = 10;
	buf[2] = buf[3] = 0;
	if (DVD_QUEUE_COMMAND(&pc)!=1) return -1; // Get Disc Key
	
	DiscKey[0] = buf[4+4];
	DiscKey[1] = buf[4+3];
	DiscKey[2] = buf[4+2];
	DiscKey[3] = buf[4+1];
	DiscKey[4] = buf[4+0];

	printf("Drive authorized!\n");
	return 0;
}

int DeCSSauth::GetStreamKey(char *FileName,CSS_Key *StreamKey)
{
	return 0;
}

int DeCSSauth::IsCiphered()
{
	return 0;
}

int DeCSSauth::DVD_QUEUE_COMMAND(packet_command *pc)
{
SRB_ExecSCSICmd MySRB;
DWORD ASPIStatus,ASPIEventStatus;

	memset(&MySRB,0,sizeof(SRB_ExecSCSICmd));
	MySRB.SRB_Cmd = SC_EXEC_SCSI_CMD;
	MySRB.SRB_HaId = Host;
	if (pc->Cmd[0]==DVD_SEND_KEY)
		MySRB.SRB_Flags = SRB_DIR_OUT | SRB_EVENT_NOTIFY;
	else
		MySRB.SRB_Flags = SRB_DIR_IN | SRB_EVENT_NOTIFY;
	MySRB.SRB_Target = Target;
	MySRB.SRB_Lun = Lun;
	MySRB.SRB_BufLen = pc->BufSize;
	MySRB.SRB_SenseLen = SENSE_LEN;
	MySRB.SRB_BufPointer = pc->Buffer;
	MySRB.SRB_CDBLen = pc->CmdSize;
	MySRB.SRB_PostProc = hEventEnd;
	memcpy(&MySRB.CDBByte,pc->Cmd,pc->CmdSize);

	ASPIStatus = SendASPI32Command((LPSRB)&MySRB);

	if (MySRB.SRB_Status == SS_PENDING)
		ASPIEventStatus = WaitForSingleObject(hEventEnd,10000);

	return MySRB.SRB_Status;
}

int DeCSSauth::IsWindowsNTor9x()
{
	if (GetVersion() < 0x80000000) // Windows NT/2000/Whistler
		return 1;
	else if (LOBYTE(LOWORD(GetVersion())) >= 4) // Windows 95/98/Me
		return 2;
	else
		return 0; // Win32s or win16
}

int DeCSSauth::GetDriveInfoNT(char DriveLetter)
{
char			DrivePath[7] = "\\\\.\\x:";
HANDLE			h;
unsigned long	bytesReturned;
SCSI_ADDRESS	dataBuffer;

	printf("Using WinNT drive-searcher\n");
	DrivePath[4] = DriveLetter;
	h = CreateFile((char*)&DrivePath,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,OPEN_EXISTING,0,NULL);
	if (!h) return 0;

	memset(&dataBuffer,0,sizeof(dataBuffer));
	if (!DeviceIoControl(h,IOCTL_SCSI_GET_ADDRESS,NULL,0,&dataBuffer,
			sizeof(dataBuffer),&bytesReturned,FALSE)) return 0;

	Host = dataBuffer.PortNumber;
	Target = dataBuffer.TargetId;
	Lun = dataBuffer.Lun;
	CloseHandle(h);
	return 1;
}

int DeCSSauth::GetDriveInfo9x(char DriveLetter)
{
int		ASPI32Status,AdapterCount,i,j,k,drive = toupper(DriveLetter)-toupper('A');
SRB_GetDiskInfo	GDI;

	printf("Using Win9x drive-searcher (%c:,%i)\n",DriveLetter,drive);
	ASPI32Status = GetASPI32SupportInfo();
	AdapterCount = (LOBYTE(LOWORD(ASPI32Status)));
	for (i=0;i<AdapterCount;i++)
		for (j=0;j<16;j++)
			for (k=0;k<8;k++)
			{
				GDI.SRB_Cmd = SC_GET_DISK_INFO;
				GDI.SRB_HaId = i;
				GDI.SRB_Flags = 0;
				GDI.SRB_Hdr_Rsvd = 0;
				GDI.SRB_Target = j;
				GDI.SRB_Lun = k;
				ASPI32Status = SendASPI32Command( (LPSRB) &GDI );
				if (ASPI32Status==1 && GDI.SRB_Int13HDriveInfo==drive)
				{
					Host = i;
					Target = j;
					Lun = k;
					return 1;
				}
			}
	Host = 0;
	Target = 0;
	Lun = 0;
	return 0;
}